/*
 * Mouse.c
 *	This should work an Optical mouse sensor using the Pololu Orangutan
 *	libraries.
 * 
 * Mr. Roboto gives this code away to do with as you will.  He does 
 * appreciate being mentioned on occasion though, a pint of bitter 
 * wouldn't be amiss either for that matter...
 */

#define ORANGUTAN

// Include our I/O definitions
#include <avr/io.h>

// Include information about the device we're building for
#include "device.h"

// Include for using delay routines
#include <util/delay.h>

// Include for the LCD subsystem
#include "lcd.h"

// Include for the ADNS5050 commands
#include "adns5050.h"

void delay_sec(unsigned char sec)
/*
 * N seconds programable delay.
 */
{
	unsigned int cycles;

	for(cycles = 0; cycles < (sec * 40); cycles ++)
	{
		_delay_ms(12);
		_delay_ms(13);
	}
}

void InitMouse(void)
/*
 * Just some basic setups
 */
{
	ACSR = 0x80; 			// turn off the comparitor
	DIDR0 = 0;				// all digital
	MCUCR &= (0<<PUD);		// Enable pull ups on pins
	InitADNS();

}

void InitLCD(void)
/*
 * Set up the Pololu 8x2 LCD display on the Orangutan
 */
{
	LED_DDR |= (1 << LED0);
	LED_PORT &= (0 << LED0);
	lcd_init();
	lcd_clear();
	lcd_hide();				// Hide cursor because it's annoying
	lcd_gotoxy(0,0);
	lcd_string("X: ");
	lcd_gotoxy(0,1);
	lcd_string("Y: ");	
}

void PrintVals(int16_t xVal, int16_t yVal)
/*
 * Print out on the display.
 * Maximum number is +/- 9999
 */
{
	lcd_gotoxy(2,0);
	if (xVal & 0x8000) {
		xVal= ~xVal;
		xVal++;
		lcd_string("-");
	}
	else {
		lcd_string(" ");
	}
	lcd_gotoxy(3,0);
	lcd_int(xVal);
	lcd_gotoxy(2,1);
	if (yVal & 0x8000) {
		yVal = ~yVal;
		yVal++;
		lcd_string("-");
	}
	else {
		lcd_string(" ");
	}
	lcd_gotoxy(3,1);
	lcd_int(yVal);	
}

int main(void)
{
	char xVal = 0;
	char yVal = 0;
	int16_t tempX, tempY;
	int16_t acumX = 0;
	int16_t acumY = 0;
	
	_delay_ms(100);	
	InitLCD();
	InitMouse();

	/*
	 * Get the Product ID (0x12) to show we're working.
	 * Get the resolution, 0 = 500 cpi, 1 = 1000 cpi
	 */
	xVal = ADNS_wr(PROD_ID,0);
	tempY = ADNS_wr(M_CTRLR,0);
	PrintVals((int16_t)xVal, (int16_t)tempY);
	_delay_ms(500);
	
	/*
	 * Sample the sensor every 10ms and accumulate the vector sum.
	 * We only have 8 bits, so if we move too fast we're likely to
	 * lose some information.
	 */
	while (1) {
		_delay_ms(10);
		xVal = ADNS_wr(MOTION,0x00);			// To trigger DELTA register reading
		xVal = ADNS_wr(DELTA_X,0x00);			// 8 bit signed value move in X axis
		yVal = ADNS_wr(DELTA_Y,0x00);			// 8 bit signed value move in Y axis
		if ((xVal != 0) || (yVal != 0)) {		// If either are != 0 the print new values
			tempX = (int16_t)xVal;
			if (xVal & 0x80) {					// negative value
				xVal = ~xVal + 1;
				tempX = (0 - (int16_t)xVal);
			}
			tempY = (int16_t)yVal;
			if (yVal & 0x80) {					// negative number
				yVal = ~yVal + 1;
				tempY = (0 - (int16_t)yVal);
			}
			acumX += tempX;
			acumY += tempY;
			PrintVals((int16_t)acumX, (int16_t)acumY);
		}
	}

	return 0;
}
